// gentrace.cc
// implementation of gentrace.h for x86-style stacks,
//   for Borland and GNU compilers

#include "gentrace.h"  // this module
#include "stackbot.h"  // stackBottom
#include "typ.h"       // bool

#include <stdio.h>     // NULL
#include <assert.h>    // assert


// The "standard" x86 stack frame, as it pertains to tracing, is quite simple:
//   dword @ [ebp]    holds caller's ebp (ebp = "Extended Base Pointer")
//   dword @ [ebp+4]  holds return address (PC in caller)
// Additionally, since the x86 stack grows towards 0, each frame is at a higher
// address as we walk back from the newest frame to the oldest.

// Barring large arrays declared on the stack, it's rare for a frame to be
// as large as 100 bytes (which would be 23 arguments and local variables
// (100/4 - 2)).  However, to allow for the possibility of reasonably small
// inline arrays, I set the largest frame size I'll believe as legitimate
// at 8 KBytes.

// As hinted above, this frame format is not always followed.  Typically,
// hand-coded or heavily optimized routines will dispense with the frame
// formalities, at the expense of the tracing routines.  To compensate, when
// the tracer runs into difficulty, it starts trying to "feel" for where the
// normal frame structure picks up again, given where it broke.

// Naturally, this kind of memory exploration is bound to try illegal addresses,
// so we require some ability to query to OS's VM manager to discover whether
// a given address will segfault.  This is the purpose of testPtr.

// During the trace, no values in the stack are modified.

// This code is intended to be called by one thread to examine the stack of
// another, if that is desired.  Obviously, the thread being examined *must*
// be suspended during the trace!


void trace(StackTrace &dest)
{
  void const *ebp;
  #if defined(__BORLANDC__)
    // Borland's handy way to grab a register value for the current thread.
    ebp = (void const*)_EBP;

  #elif defined(__GNUC__)
    // GNU provides a similar mechanism
    ebp = (void const*)__builtin_frame_address(0);

  #else
    // address-of dummy variable is a possibility...
    #error I dont know how to get frame pointer for this compiler
  #endif

  trace(dest, ebp, NULL);
}


// define some parameters for the trace
enum {
  TRAIL_DEPTH = 2,
    // For each candidate trail start, we'll follow it this many steps, checking
    // each for being an apparently valid frame, before deciding it's the right
    // one.  I've observed incorrect paths with 1 valid hop and correct paths
    // with only 2 valid hops before the next break, so unless I get a lot
    // more sophisticated I have no choice on this number.

  MAX_FRAME_SIZE = 0x2000,       // 8kbytes
    // This is largest frame size we'll believe is real.

  HOW_FAR_BACK = MAX_FRAME_SIZE,
    // Once the chain appears broken, we'll examine this number / 4 entries
    // looking for the where the path begins again.  Now that I've made the
    // valid-frame test more stringent, we can and often must examime many
    // more entries.
};


// I'll leave this commented-out until/if I think it is actually needed
//#define Frame X86StackFrame       // ensure no linker collisions


bool inStack(void const *ptr, void const *stackTop)
{
  // x86 stacks grow towards 0
  return stackBottom > ptr &&
                       ptr >= stackTop;
}



// define the format of a frame
struct Frame {
  Frame const *next;       // next frame
  void const *eip;         // return address

  unsigned size() const { return (int)next - (int)this; }
    // size of the frame, based on where the previous frame begins

  bool isLastFrame() const { return next == 0 || eip == 0; }
    // note that the last frame is *not* reported as valid

  bool isValidFrame(void const *stackTop) const;
    // true if this frame seems ok
};


// I had some stuff here about the exception stack; it's now down
// in the 'trash' section because I hope I won't need it...


// test to see if a given frame pointer looks like it points at a valid frame
// BUT not the last frame
bool Frame::isValidFrame(void const *stackTop) const
{
  return inStack(&next, stackTop) &&
         inStack(&eip, stackTop) &&
           // both parts of the frame must be accessible, otherwise we'd
           // segfault when we try to access next or eip!
         eip != 0 &&
           // need a real return address
         next > this &&
           // must point to older frame
         inStack(next, stackTop) &&
           // older frame is accessible
         size() < MAX_FRAME_SIZE;
           // frame size must be within limits
}


// to have something to do where I want to be able to set a breakpoint
static void foo() {}


void StackTrace::addAddr(void const *eip)
{
  if (next < length) {
    trace[next++] = eip;
  }
  else {
    // do nothing; no room
  }
}


void trace(StackTrace &dest, void const *FP, void const *PC)
{
  // make sure initStackBottom has been called
  if (stackBottom == NULL) {
    // for complex programs, we often see calls to 'new' before main
    // even begins (because of global object ctors); so let's just do
    // the init now
    initStackBottom();
  }

  // use x86 name for a frame pointer ("base pointer", e="extended"=32 bits)
  Frame const *ebp = (Frame const*)FP;

  // add the first eip address to caller stack record
  if (PC) {
    // if PC is zero, caller didn't want it to be part of the trace; if it
    // was nonzero, then it will be (this may seem trivial, but when the
    // trace is generated by another thread, failing to record this PC fails
    // to record the profiling information which may well be the central
    // purpose!)
    dest.addAddr(PC);
  }

  // limit here is simply to catch infinite loop; an inf loop should
  // be prevented by the fact that a frame is only considered valid
  // if it points to an older frame
  int iters;    enum { MAX_ITERS=10000 };
  for(iters=0; iters<MAX_ITERS; iters++) {
    // test accessibility of next frame
    if (!ebp->isValidFrame(FP)) {
      // ebp doesn't appear to point anywhere useful

      // see if it's something I'll accept as the real last entry
      if (ebp->isLastFrame()) {
        foo();     // so can set a breakpoint
        break;
      }

      // this looks like a nonstandard stack frame.. let's see if we
      // can pick up the trail again!
      int startOfs;
      for(startOfs = 0; startOfs < HOW_FAR_BACK;
          startOfs += sizeof(void*)) {
        Frame const *frame = (Frame const *)((int)ebp + startOfs);
        int i;
        for (i=0; i<TRAIL_DEPTH; i++) {
          if (frame->isValidFrame(FP)) {
            // ok! go to next one
            frame = frame->next;
              // during this exploration I don't track eip; if the trail ends
              // up being followed, I'll be grabbing eip on the second pass
          }
          else {
            // trail dead-ends
            break;
          }
        }

        if (i==TRAIL_DEPTH) {
          // looks like we're back on the trail! let's reset frame
          // and ebp and get back to business
          ebp = (Frame const *)((int)ebp + startOfs);
          break;
        }
      }

      if (startOfs == HOW_FAR_BACK) {
        // couldn't find the trail again.. give up
        foo();     // so can set a breakpoint
        break;
      }
    }

    // at this point, we are convinced that ebp points to
    // a valid frame (which implies, among other things, that
    // it is not the last frame)

    // store in trace
    dest.addAddr(ebp->eip);

    // use frame pointer to get next ebp
    ebp = ebp->next;

    // process next frame
  }

  assert(iters != MAX_ITERS);
}


// -------------- test code -------------------------
#ifdef TEST_GENTRACE

// hypothetical user of pointers
void useIt(...);

// this macro causes the compiler to *not* optimize-away the
// normal stack frame structure
#define FORCE_FRAME int dummy; useIt(&dummy);

// chain of callers
void calledFirst();
void calledSecond();
void calledThird();
void printTrace();    // last one called

void calledFirst()
{
  FORCE_FRAME
  calledSecond();
}

void calledSecond()
{
  FORCE_FRAME
  calledThird();
}

void calledThird()
{
  FORCE_FRAME
  printTrace();
}

void printTrace()
{
  // create the trace structure
  StackTrace tr;
  enum { TRACE_LEN = 100 };
  tr.trace = new void const *[TRACE_LEN];
  tr.length = TRACE_LEN;
  tr.next = 0;

  // run the tracer
  trace(tr);

  // print the results (first thing printed is most-recent fn)
  for (int i=0; i<tr.next; i++) {
    printf("%p\n", tr.trace[i]);
  }
}

void useIt(...)
{}


int main()
{
  FORCE_FRAME
  initStackBottom();
  calledFirst();
  return 0;
}


#endif // TEST_GENTRACE


#if 0       // trash

// Isn't this an interesting twist: the Borland compiler already knows about
// _ExceptionHandler, though it's not mentioned in *any* header file.
// Furthermore, the signature given below doesn't match what the compiler
// thinks it is, and the compiler complains.  But, since the compiler already
// knows, there's no need to delcare it at all!

#if 0
// we'll need to know this runtime address
extern "C" {      // the linker knows it as "__ExceptionHandler"
  extern void _ExceptionHandler();
    // the actual arguments and return value are unknown but irrelevant for
    // this module's purposes
};
#endif // 0

         testPtr(eip) &&
           // return address must point to accessible memory (otherwise it's
           // obviously not a real return address)
         eip != _ExceptionHandler;
           // (Roughly) parallel to the main frame chain, there is a chain of
           // exception handlers, which has pointers to the handler functions
           // exactly where we're expecting return addresses!  The only way
           // to distinguish this chain from the one we want is the fact that
           // most (all?) of the handler pointers point to this one specific
           // function.


  MAX_TRACE_LENGTH = 300,
    // the fact that testing a pointer is slow forces me to stop the trace
    // prematurely if it gets too big; on my 486-33, it can test about 300
    // frames per second, so this gives about a second of tracing; I considered
    // having the routine periodically check a clock, but then that has to
    // be disabled for debugging, which is a hassle
#endif // 0   (trash)

