// mapfile.cc
// code for mapfile.h

// expected file format:
//   <hex-start-addr> <name>
// in order of increasing address


#include "mapfile.h"     // this module
#include "xassert.h"     // xassert

#include <stdio.h>       // FILE
#include <stdlib.h>      // strtoul
#include <string.h>      // strchr 

#ifdef USE_EXCEPTIONS
  #include "exc.h"         // XOpen
  #include "syserr.h"      // xsyserror
#else
  #define xopen(fn)                                          \
    fprintf(stderr, "MapFile: cannot open %s\n", fn);        \
    entries = NULL;                                          \
    return
  #define xsyserror(fn)  \
    perror(fn);          \
    entries = NULL;      \
    return
#endif


MapFile::MapFile(char const *fname)
{
  FILE *fp = fopen(fname, "rt");
  if (!fp) {
    xopen(fname);    // throws exception
  }

  // count lines in the file
  enum { BUFLEN = 256 };
  char buf[BUFLEN];
  numEntries=0;
  while (fgets(buf, BUFLEN, fp)) {
    numEntries++;
  }

  if (fseek(fp, 0, SEEK_SET) != 0) {
    xsyserror("fseek");
  }

  // allocate array
  entries = new Entry[numEntries];

  // scan file again, entering info into the file
  int index=0;
  while (fgets(buf, BUFLEN, fp)) {
    // make sure things are still as expected
    xassert(index < numEntries);

    // parse address
    entries[index].addr = (void const*)
      strtoul(buf, NULL, 16 /*radix*/);

    // find separating whitespace
    char const *ws = strchr(buf, ' ');
    char const *nl = strchr(buf, '\n');
    if (ws == NULL || nl == NULL) {
      fprintf(stderr, "%s: line %d is malformed\n", fname, index+1);
      continue;
    }
    ws++;    // advance to name's first char

    // grab name
    entries[index].name = string(ws, nl-ws);

    // go to next entry
    index++;
  }

  // should have exactly filled array
  xassert(index == numEntries);

  // close file
  if (fclose(fp) != 0) {
    delete[] entries;
    xsyserror("fclose");
  }
}


MapFile::~MapFile()
{
  // automatically deallocates names
  delete[] entries;
}


char const *MapFile::map(void const *addr)
{
  // check for initialization error
  if (!entries) {
    return NULL;
  }

  // check for too-small
  if (addr < entries[0].addr) {
    return NULL;
  }

  // binary search to find the right entry
  int low = 0;
  int high = numEntries-1;

  while (low < high) {
    // compare to middle entry
    int mid = (low+high+1)/2;   // +1 to round up; mid > low always
    xassert(mid > low);         // guarantee progress
    if (addr < entries[mid].addr) {
      // target is in low half
      high = mid-1;       // can't be 'mid' itself
    }
    else {
      // target is in high half
      low = mid;          // *can* be 'mid' itself
    }
  }

  xassert(low == high);
  return entries[low].name.pcharc();
}


// ---------------- test code ------------------
#ifdef TEST_MAPFILE

int main(int argc, char *argv[])
{
  if (argc < 2) {
    printf("usage: %s mapfile.map\n"
           "  input line format:\n"
           "    <hexadecimal-address>\n"
           "  output line format:\n"
           "    <symbol-name>\n", argv[0]);
    return 0;
  }

  // read map
  MapFile map(argv[1]);

  char buf[80];
  while (fgets(buf, 80, stdin)) {
    int addr = strtoul(buf, NULL, 16 /*radix*/);
    char const *name = map.map((void const*)addr);
    if (name != NULL) {
      printf("%s\n", name);
    }
    else {
      printf("(address too low)\n");
    }
  }

  return 0;
}

#endif // TEST_MAPFILE

